#!/usr/bin/perl

use mapscript410;
use CGI qw(:standard :html escape);
use IO::String;

# use strict;

my $timestamp = $$ . time;

sub error() {
  my($message) = @_;

  # restore original filehandle, output error as XML and exit
  select($main::orig_fh);
  print $main::cgi->header('text/xml');
  print "<dLegend>\n";
  print "  <error>\n";
  print "    $message\n";
  print "  </error>\n";
  print "</dLegend>\n";

  exit 1;
}

sub min {
		if ($_[0]<$_[1]) {return $_[0]} else {return $_[1]};
}

sub max {
		if ($_[0]<$_[1]) {return $_[1]} else {return $_[0]};
}

sub get_layer_icon() {
		my( $map, $layer) = @_;		
			
		my $icon;
		for (my $i=0; $i<$layer->{numclasses}; $i++) {
				my $class = $layer->getClass($i);

				# skip raster layer classes if there is no key image
				next if ($layer->{type} == $mapscript410::MS_LAYER_RASTER && !($class->{keyimage}));

				if ($class->{name}) {
						$icon = $class->createLegendIcon($map, $layer, $map->{legend}->{keysizex}, $map->{legend}->{keysizey}) or &error(mapscript410::msGetErrorString(','));
						last;
				}
		}

		return '' if !($icon);

		my $extension;
		if ($map->{imagetype} =~ /^png|^PNG/) {
				$extension = '.png';
		} elsif ($map->{imagetype} =~ /^jpeg|^JPEG/) {
				$extension = '.jpg';
		} else {
				$extension = '.gif';
		}

		my $filename = 'dlegend_'. $timestamp .'_'. $layer->{name} .$extension;

		$icon->save($map->{web}->{imagepath} ."$filename", $map);

		return $map->{web}->{imageurl} ."$filename";
}

sub get_class_icon() {
		my( $map, $layer, $class, $class_index) = @_;		
			
		my $icon;		
		if ($class->{name}) {
				$icon = $class->createLegendIcon($map, $layer, $map->{legend}->{keysizex}, $map->{legend}->{keysizey}) or &error(mapscript410::msGetErrorString(','));
		}						

		return '' if !($icon);

		my $extension;
		if ($map->{imagetype} =~ /^png|^PNG/) {
				$extension = '.png';
		} elsif ($map->{imagetype} =~ /^jpeg|^JPEG/) {
				$extension = '.jpg';
		} else {
				$extension = '.gif';
		}

		my $filename = 'dlegend_'. $timestamp .'_'. $layer->{name} .'_'. $class_index .$extension;

		$icon->save($map->{web}->{imagepath} ."$filename", $map);

		return $map->{web}->{imageurl} ."$filename";
}

sub get_layer_name() {
		my($metadata_value, $layer, $number_of_classes) = @_;		
		
		return $metadata_value if $metadata_value;
		
		return undef if $layer->{numclasses} != 1;
		
		for(my $i=0; $i<$layer->{numclasses}; $i++) {
				my $class = $layer->getClass($i);
				return $class->{name} if $class->{name};
		}
		
		return undef; # shouldn't get here
}

sub number_of_valid_classes() {
		my ($layer) = @_;
		
		my $n = 0;
		for(my $i=0; $i<$layer->{numclasses}; $i++) {
				my $class = $layer->getClass($i);
				$n++ if $class->{name};
		}

		return $n;
}

sub is_layer() {
		my ($name, $map) = @_;

		my $layer = $map->getLayerByName($name);
		return ($layer) ? 1:0;		
}

sub get_group_minscale_and_maxscale() {
		my ($map, @layer_indexes) = @_;
		
		my $minscale = -1; 
		my $maxscale = -1;
		my $has_no_minscale = 0;
		my $has_no_maxscale = 0;
		for (my $i=0; $i<=$#layer_indexes; $i++) {
				my $layer = $map->getLayer($layer_indexes[$i]);		

				$has_no_minscale = 1 if $layer->{minscale} == -1;
				$has_no_maxscale = 1 if $layer->{maxscale} == -1;

				if ($layer->{minscale} != -1) {
						$minscale = ($minscale == -1) ? $layer->{minscale} : &min($minscale, $layer->{minscale});
				}

				if ($layer->{maxscale} != -1) {
						$maxscale = ($maxscale == -1) ? $layer->{maxscale} : &max($maxscale, $layer->{maxscale});
				}
		}
	
		$minscale = -1 if $has_no_minscale;
		$maxscale = -1 if $has_no_maxscale;

		return ($minscale, $maxscale);
}

sub layers_in_group() {
		my ($name, $map) = @_;
		my @list;

		for(my $i=0; $i<$map->{numlayers}; $i++) {
				my $layer = $map->getLayer($i);
				push @list, $i if $layer->{group} eq $name;
		}

		return @list;
}

sub write_layer() {
		my($layer, $map, $tab) = @_;

		my $number_of_classes = &number_of_valid_classes($layer);
		my $name = &get_layer_name($map->{web}->{metadata}->get("dlegend_". $layer->{name} ."_name"), $layer, $number_of_classes);
		
		if ($number_of_classes <= 1) {
				my $icon = &get_layer_icon($map, $layer);

        return if !($name);

				print "$tab<layer id=\"". $layer->{name} ."\" name=\"$name\"";
				print " minscale=\"". $layer->{minscale} ."\"" if $layer->{minscale} > 0;
				print " maxscale=\"". $layer->{maxscale} ."\"" if $layer->{maxscale} > 0;
				print " icon=\"$icon\"" if $icon;
				print " />\n";
		} else {												

				if($name) {
						print "$tab<layer id=\"". $layer->{name} ."\" name=\"$name\"";
						print " minscale=\"". $layer->{minscale} ."\"" if $layer->{minscale} > 0;
						print " maxscale=\"". $layer->{maxscale} ."\"" if $layer->{maxscale} > 0;
						print " >\n";
				}
				
				for (my $i=0; $i<$layer->{numclasses}; $i++) {
						my $class = $layer->getClass($i);
						my $icon = &get_class_icon($map, $layer, $class, $i);
						if ($class->{name}) {
								if($name) {
										print "$tab  <class name=\"". $class->{name} ."\"";
										print " icon=\"$icon\"" if $icon;
										print " />\n";
								} else {
										# promote the class to a layer for display (this should only happen within a group)
										print "$tab<layer id=\"". $layer->{name} ."\" name=\"". $class->{name} ."\"";
										print " minscale=\"". $layer->{minscale} ."\"" if $layer->{minscale} > 0;
										print " maxscale=\"". $layer->{maxscale} ."\"" if $layer->{maxscale} > 0;
										print " icon=\"$icon\"" if $icon;
										print " />\n";
								}
						}
				} 

				print "$tab</layer>\n" if $name;
		}
}

#
# main script
#

# give explicit package name for access in error()
$main::cgi = new CGI;

# set a string buffer to print to; store original filehandle (stdout)
# in $orig_fh
my $string_buffer;
my $string_fh = IO::String->new($string_buffer);

# give explicit package name for access in error()
$main::orig_fh = select($string_fh);

print $cgi->header('text/xml');
&error("Parameter 'map' is not set.") if !($cgi->param('map'));
&error("Parameter 'server' is not set.") if !($cgi->param('server'));

my $mapfile = $cgi->param('map');
my $server = $cgi->param('server');
my $map = new mapscript410::mapObj($mapfile) or &error(mapscript410::msGetErrorString(','));

# start the XML output
print "<?xml version=\"1.0\" ?>\n";
print "<dLegend>\n";

print "  <server>$server</server>\n";

# dlegend_metagroups
my $value = $map->{web}->{metadata}->get("dlegend_metagroups");
if ($value) {
		my @metagroups = split ',', $value;
    foreach (@metagroups) {
				print "  <metagroup";				
				print " name=\"$value\"" if ($value = $map->{web}->{metadata}->get("dlegend_". $_ ."_name"));
				print " element=\"$value\"" if ($value = $map->{web}->{metadata}->get("dlegend_". $_ ."_element"));
				print ">\n";

				$value = $map->{web}->{metadata}->get("dlegend_". $_ ."_layers_and_groups");
				if ($value) {
						my @layers_and_groups = split ',', $value;
						foreach (@layers_and_groups) {
								if (&is_layer($_, $map)) {
										my $layer = $map->getLayerByName($_);
										&write_layer($layer, $map, '    ');
								} else {
										my @layer_indexes = &layers_in_group($_, $map);
										
										$value = $map->{web}->{metadata}->get("dlegend_". $_ ."_name");										
										my $name = ($value) ? $value : undef;

										my ($minscale, $maxscale) = &get_group_minscale_and_maxscale($map, @layer_indexes);
										
										if($name) { 
												print "    <group id=\"$_\" name=\"$name\"";

												print " minscale=\"$minscale\"" if $minscale > 0;
												print " maxscale=\"$maxscale\"" if $maxscale > 0;
												
												print " >\n";
												
												for (my $i=0; $i<=$#layer_indexes; $i++) {
														my $layer = $map->getLayer($layer_indexes[$i]);
														&write_layer($layer, $map, '      ');
												}
												
												print "    </group>\n";
										} else  {
												# promote the group to a layer (don't write a group)
												for (my $i=0; $i<=$#layer_indexes; $i++) {
														my $layer = $map->getLayer($layer_indexes[$i]);

														my $number_of_classes = &number_of_valid_classes($layer);
														if($number_of_classes > 0) {																

																# alter a few layer properties based on the group
																$layer->{name} = $_;
																$layer->{minscale} = $minscale;
																$layer->{maxscale} = $maxscale;

																&write_layer($layer, $map, '    ');
														}
												}
										}
								}
						}
				}

				print "  </metagroup>\n";
    }
} else {
  # TODO: deal with a legend without metagroups
}

print "</dLegend>\n";

# restore original filehandle and print the string buffer
select($orig_fh);
print $string_buffer;

# done
exit 0;
