#!/usr/bin/python -u

import getopt
import mapscript
import os
import re
import sys
from StringIO import StringIO
import time

__doc__="""
XML legend generator (for consumption by dlegend.js)

Usage: python make_dlegend_xml.py [params]

Params:
  -h, --help                      print this help
  -m <mapfile>, --map=<mapfile>   full path to .map file
  -s <server>, --server=<server>  server name (less protocol)

Examples:
  python make_dlegend_xml.py -m /usr/local/www/docs/app/app.map -s my.server.com
  python make_dlegend_xml.py --map=/usr/local/www/docs/app/app.map --server=my.server.com

"""

timestamp = "%s%s" % (os.getpid(), int(time.time()))

def usage():
    print __doc__


def error(handle):
    # empty out the handle we're writing to, then insert the error text
    handle.seek(0)
    handle.truncate()
    print "<dLegend>\n  <error>"
    (type, value) = sys.exc_info()[:2]
    print "    %s:\n    %s" % (type, value)
    print "  </error>\n</dLegend>"


def is_layer(name, mapp):
    layer = mapp.getLayerByName(name)
    if layer:
        return 1
    else:
        return 0


def write_layer(layer, mapp, tab):
    number_of_classes = number_of_valid_classes(layer)
    name = get_layer_name( \
        mapp.web.metadata.get('dlegend_'+layer.name+'_name'), \
        layer, number_of_classes) 

    if number_of_classes <= 1:

        icon = get_icon(mapp, layer)

        # Why is this here?
        if not name:
            return

        print '%s<layer id="%s" name="%s"' % (tab, layer.name, name),
        if layer.minscale > 0:
            print 'minscale="%s"' % (layer.minscale),
        if layer.maxscale > 0:
            print 'maxscale="%s"' % (layer.maxscale),
        if icon:
            print 'icon="%s"' % (icon),
        print '/>'

    else:
        
        if name:
            print '%s<layer id="%s" name="%s"' % (tab, layer.name, name),
            if layer.minscale > 0:
                print 'minscale="%s"' % (layer.minscale),
            if layer.maxscale > 0:
                print 'maxscale="%s"' % (layer.maxscale),
            print '>'
            
        for i in range(0, layer.numclasses):
            klass = layer.getClass(i)
            icon = get_icon(mapp, layer, klass, i)
            if klass.name:
                if name:
                    print '%s  <class name="%s"' % (tab, klass.name),
                    if icon:
                        print 'icon="%s"' % (icon),
                    print '/>'
                else:
                    # promote the class to a layer for display (this should only happen within a group)
                    print '%s<layer id="%s" name="%s"' % (tab, layer.name, klass.name),
                    if layer.minscale > 0:
                        print 'minscale="%s"' % (layer.minscale),
                    if layer.maxscale > 0:
                        print 'maxscale="%s"' % (layer.maxscale),
                    if icon:
                        print 'icon="%s"' % (icon),
                    print '/>'

        if name:
            print '%s</layer>' % (tab)
            
    
def get_icon(mapp, layer, klass=None, klass_index=None):

    icon = None

    if not klass:
        for i in range(0, layer.numclasses):
            klass = layer.getClass(i)

            # skip raster layer classes if there is no key image
            if layer.type == mapscript.MS_LAYER_RASTER and not klass.keyimage:
                continue

            if klass.name:
                # it has a name, this is the one we want
                break
                # next test appears redundant; but is needed for case
                # where klass is passed directly
        
    if klass.name:
        icon = klass.createLegendIcon( \
            mapp, layer, mapp.legend.keysizex, mapp.legend.keysizey)

        # This bit appears to be unnecessary; current exception
        # handling should cover it.  Needs testing.  Else create 
        # MapServer exception and raise it.        
##        if not icon:
##            print mapscript.msGetErrorString(',')

    if not icon:
        return ''

    extension = ''
    if re.match('png', mapp.imagetype, re.I):
        extension = '.png'
    elif re.match('jpg', mapp.imagetype, re.I):
        extension = '.jpg'
    else:
        extension = '.gif'
        
    filename = 'dlegend_' + timestamp + '_' + layer.name
    if not klass_index == None:
        filename = filename + str(klass_index)
    filename = filename + extension

    # exceptions here (if, eg., mapfile WEB IMAGEPATH is invalid) for
    # whatever reason do not bubble up to the try/catch block in
    # process_args(); traceback is output immediately.  May need to
    # catch here and send directly to error(), which will mean
    # reworking logic there and in process_args().
    #
    icon.save(mapp.web.imagepath + filename, mapp)
    
    return mapp.web.imageurl + filename
    
    
def layers_in_group(name, mapp):
    indexes = []

    for i in range(0, mapp.numlayers):
        layer = mapp.getLayer(i)
        if layer.group == name:
            indexes.append(i)

    return indexes
    

def get_layer_name(metadata_value, layer, number_of_classes):
    if metadata_value:
        return metadata_value

    if number_of_classes != 1:
        return None

    for i in range(0, layer.numclasses):
        klass = layer.getClass(i)
        if klass.name:
            return klass.name

    return None     # shouldn't reach this point


def number_of_valid_classes(layer):
    n = 0
    for i in range(0, layer.numclasses):
        klass = layer.getClass(i)
        if klass.name:
            n += 1
    return n


def get_group_minscale_and_maxscale(mapp, layer_indexes):
    g_minscale = -1
    g_maxscale = -1
    g_has_no_minscale = False
    g_has_no_maxscale = False
    for i in range(0, len(layer_indexes)):
        layer = mapp.getLayer(layer_indexes[i])

        if layer.minscale == -1:
            g_has_no_minscale = True
        else:
            if g_minscale == -1:
                g_minscale = layer.minscale
            else:
                g_minscale = min(g_minscale, layer.minscale)

        if layer.maxscale == -1:
            g_has_no_maxscale = True
        else:
            if g_maxscale == -1:
                g_maxscale = layer.maxscale
            else:
                g_maxscale = max(g_maxscale, layer.maxscale)

    if g_has_no_minscale:
        g_minscale = -1
    if g_has_no_maxscale:
        g_maxscale = -1

    return (g_minscale, g_maxscale)


def write_xml(mapfile, server):
    
    mapp = mapscript.mapObj(mapfile)

    # start XML output
    print '<?xml version="1.0" ?>'
    print '<dLegend>'
    print '  <server>%s</server>' % server

    # dlegend_metagroups
    value = mapp.web.metadata.get('dlegend_metagroups')
    if value:
        metagroups = value.split(',')
        for metagroup in metagroups:

            # metagroup opening tag
            print '  <metagroup',
            value = mapp.web.metadata.get('dlegend_' + metagroup + '_name')
            if value:
                print 'name="%s"' % value,
            value = mapp.web.metadata.get('dlegend_' + metagroup + '_element')
            if value:
                print 'element="%s"' % value,
            print '>'

            # layers and groups within metagroup
            value = mapp.web.metadata.get('dlegend_' + metagroup + '_layers_and_groups')
            if value:
                layers_and_groups = value.split(',')

                # l_g is either layer or group
                for l_g in layers_and_groups:

                    # just write out the layer info if it's a layer
                    if is_layer(l_g, mapp):
                        layer = mapp.getLayerByName(l_g)
                        write_layer(layer, mapp, '    ')

                    # handling for group
                    else:
                        layer_indexes = layers_in_group(l_g, mapp)
#                        print "%s LI: %s" % (l_g, layer_indexes)

                        # get group name defined in web metadata, if there is one
                        value = mapp.web.metadata.get('dlegend_'+l_g+'_name')
                        if value:
                            name = value
                        else:
                            name = None

                        (minscale, maxscale) = get_group_minscale_and_maxscale(mapp, layer_indexes)

                        # write group tag if name was spec'd
                        if name:
                            print '    <group id="%s" name="%s"' % (l_g, name),
                            if minscale > 0:
                                print 'minscale="%s"' % (minscale),
                            if maxscale > 0:
                                print 'maxscale="%s"' % (maxscale),
                            print '>'

                            # write layer tags for layers within group
                            for i in range(0, len(layer_indexes)):
                                 layer = mapp.getLayer(layer_indexes[i])
                                 write_layer(layer, mapp, '      ')

                            print '    </group>'

                        # treat group as layer(s) if no name was spec'd in web metadata
                        # (likely scale-swappable group of layers)
                        else:

                            for i in range(0, len(layer_indexes)):
                                layer = mapp.getLayer(layer_indexes[i])
                                number_of_classes = number_of_valid_classes(layer)

                                if number_of_classes > 0:

                                    # alter a few layer properties based on the group
                                    layer.name = l_g
                                    layer.minscale = minscale
                                    layer.maxscale = maxscale

                                    write_layer(layer, mapp, '    ')


            print '  </metagroup>'

    else:
        # TODO: deal with a legend without metagroups
        thismeans = 'nothing' # python expects intended block here

    print '</dLegend>'

    return 1


def process_args(argv):
    mapfile = None
    server = None
    try:
        opts, args = getopt.getopt(argv, "hm:s:", ["help", "map=", "server="])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit()
        elif opt in ('-m', '--map'):
            mapfile = arg
        elif opt in ('-s', '--server'):
            server = arg
    if not mapfile or not server:
        # for command-line only.  make_dlegend_xml_cgi.py handles this on its own.
        usage()
        sys.exit(1)
    else:
        stdout = sys.stdout
        sys.stdout = sio = StringIO()
        try:
            write_xml(mapfile, server)
        except:
            error(sio)
        sys.stdout = stdout
        print sio.getvalue()
            

if __name__ == '__main__':
    process_args(sys.argv[1:])
